

         /*********************************************************/
         /*                                                       */
         /*                 COPYRIGHT (C) 1988 BY                 */
         /* CAD Group of Dipartimento di Automatica e Informatica */
         /*             Politecnico di Torino, Italy.             */
         /*                  ALL RIGHTS RESERVED.                 */
         /*                                                       */
         /* This software is furnished under a license and may be */
         /* used and copied  only in accordance with the terms of */
         /* such license. Unauthorized copies will be persecuted. */
         /*                                                       */
         /*********************************************************/





/******************************************************************************

    Filename:         crefl.c

    Author:           Pier Luca Montessoro

    Creation date:    16 January 1989

    Contents:         fault list translator


                          *** DOCUMENTATION ***


 ******************************************************************************/



#include "env.h"
#include "common.h"
#include "loadcode.h"
#include "topology.h"


#define  LINKER_VERSION       "3.0"



/* topological variables */

/* delay table element structure definition */

typedef struct Delay_Table_Element
        {
           unsigned long  rise;
           unsigned long  fall;
        }  DELAY_TABLE_ELEMENT;


DELAY_TABLE_ELEMENT    *delay_table;
TOPOLOGICAL_ELEMENT    *topology_vector;
FANIN_ELEMENT          *fanin_vector;
FANOUT_ELEMENT         *fanout_vector;

unsigned long          num_of_elements;
unsigned long          max_level_number;
unsigned long          max_eval_link_number;




#define  MAX_PERIOD_LENGTH     64

#define  IN_FL_EXTENSION    "fau"
#define  OUT_FL_EXTENSION   "fls"



/* network file */

FILE                   *net_file;


/* input fault list file */
                     
FILE                   *in_fl_file;


/* output fault list file */

FILE                   *out_fl_file;




/******************************************************************************

    Function name:    main

    Authors:          Pier Luca Montessoro

    Creation date:    21 June 1988

    Purpose:          main

 ******************************************************************************/


main (argc, argv)

int   argc;
char  *argv[];

{

   /* local variables */

   char                  net_file_name [MAX_FILE_NAME_LENGTH];
   char                  in_fl_file_name [MAX_FILE_NAME_LENGTH];
   char                  out_fl_file_name [MAX_FILE_NAME_LENGTH];


   /* beginning of executable part */

   if (argc != 3)
   {
      printf ("Wrong number of arguments (1st: fault list, 2nd: net list).\n");
      good_bye (FALSE, TRUE);
   }


   /* open the net, source and the output files */

   strcpy (net_file_name, argv[2]);
   strcat (net_file_name, ".");
   strcat (net_file_name, NET_EXTENSION);

   strcpy (in_fl_file_name, argv[1]);
   strcat (in_fl_file_name, ".");
   strcat (in_fl_file_name, IN_FL_EXTENSION);

   strcpy (out_fl_file_name, argv[1]);
   strcat (out_fl_file_name, ".");
   strcat (out_fl_file_name, OUT_FL_EXTENSION);

   if ((in_fl_file = fopen (in_fl_file_name, "r")) == NULL)
   {
      printf ("I cannot open the input file.\n");
      good_bye (TRUE, TRUE);
   }

   if ((out_fl_file = fopen (out_fl_file_name, "w")) == NULL)
   {
      printf ("I cannot open the output file.\n");
      good_bye (TRUE, TRUE);
   }

   if ((net_file = fopen (net_file_name, "r")) == NULL)
   {
      printf ("I cannot open the network file.\n");
      good_bye (FALSE, TRUE);
   }

   load_net ();

   fau_gen ();

   printf ("Output file %s written.\n", out_fl_file_name);

   good_bye (TRUE, FALSE);

}  /* end of main */



/******************************************************************************

    Function name:    fau_gen

    Authors:          Pier Luca Montessoro

    Creation date:    16 January 1989

    Purpose:          to translate fault list

 ******************************************************************************/


fau_gen ()

{

   /* local variables */

   char                 release[MAX_MESSAGE_LENGTH];
   long                 fault_desc;
   long                 index;
   long                 dummy1;
   long                 dummy2;
   long                 pin;
   long                 inp_index;
   long                 bit = 0;
   long                 value;


   /* beginning of executable part */

   fscanf (in_fl_file, "%s", release);


   /* check the fault manager version */

   if (strcmp (release, LINKER_VERSION))
   {
      printf ("The fault manager version is not compatible with me.\n");
      good_bye (TRUE, TRUE);
   }


   while (fscanf (in_fl_file, "%ld %ld %ld %ld",
                  &fault_desc, &index, &dummy1, &dummy2) != EOF)
   {
      pin = fault_desc & 127;

      if (pin != 0)
      {
         switch (topology_vector[index].type)
         {
            case DFF_PET: 
            case TFF_PET: 
            case DFF_NET:
            case TFF_NET:
                    pin = pin - 2;
                    break;
   
            case DFF_PET_C:
            case TFF_PET_C:
            case DFF_NET_C:
            case TFF_NET_C:
                    pin = pin - 1;
                    break;
   
            case DFF_PET_S:
            case TFF_PET_S:
            case DFF_NET_S:
            case TFF_NET_S:
                    if (pin > 1)
                    {
                       pin = pin - 1;
                    }
                    break;
   
            default: /* no change */
                    break;
         }

         /* compute the index of the element connected to the fault 
            input */

         inp_index = ((topology_vector[index].fanin_position + pin - 1)->link - 
                      topology_vector);

      }
      else
      {
         inp_index = (long) -1;
      }

      value = (fault_desc & ~127) >> 7;

      fprintf (out_fl_file, "%ld %ld %ld %ld\n",
               value, inp_index, bit, index);

   }  /* end of reading the fault description file */


   return;

}  /* end of fau_gen */



/******************************************************************************

    Function name:    good_bye

    Authors:          Pier Luca Montessoro

    Creation date:    21 June 1988

    Purpose:          exit function

 ******************************************************************************/


good_bye (open_files, error_found)

unsigned char  open_files;
unsigned char  error_found;

{

   /* beginning of executable part */

   if (open_files)
   {
      fclose (net_file);
      fclose (in_fl_file);
      fclose (out_fl_file);
   }

   if (error_found)
   {
      printf ("I am sorry, but I cannot proceed. Try again.\n");
   }
   else
   {
      printf ("Job terminated. Good-bye!\n");
   }

   if (error_found)
   {
      exit (EXIT_ER);
   }
   else
   {
      exit (EXIT_OK);
   }


}  /* end of good_bye */



/******************************************************************************

    Function name:    get_memory

    Authors:          Pier Luca Montessoro

    Creation date:    25 January 1988

    Purpose:          memory allocation for network data structures

 ******************************************************************************/


char  *get_memory (number, size)

unsigned long  number;
unsigned long  size;

{

   /* local variable */

   char   *alloc_pointer;


   /* beginning of executable part */

   alloc_pointer = (char *) malloc (number * size);


   if (alloc_pointer == NULL)
   {
      printf ("No memory is available for topological data allocation.\n");
      good_bye (TRUE, TRUE);
   }

   return (alloc_pointer);

}  /* end of get_memory */



/******************************************************************************

    Function name:    load_net

    Authors:          Pier Luca Montessoro

    Creation date:    25 January 1988

    Purpose:          network data structure creation

 ******************************************************************************/


load_net ()

{

   /* called function */

   char                  *get_memory();


   /* local variables */

   char                  net_file_name [MAX_FILE_NAME_LENGTH];
   TOPOLOGICAL_ELEMENT   *second_pass = (TOPOLOGICAL_ELEMENT *) NULL;


   /* general information variables */

   char              linker_version [MAX_NAME_LENGTH];
   char              circuit_name   [MAX_NAME_LENGTH];


   /* symbol table and module table offsets */

   long int          sym_tab_offset;
   long int          mod_tab_offset;


   /* number of warnings */

   unsigned long     warnings;


   /* information related to the topological elements */

   unsigned long     p_o_num;
   unsigned long     p_i_num;
   unsigned long     p_b_num;
   unsigned long     p_o_size;
   unsigned long     p_i_size;
   unsigned long     p_b_size;
   unsigned long     constants_num;
   unsigned long     nodes_num;
   unsigned long     wires_num;
   unsigned long     regi_num;
   unsigned long     wires_size;
   unsigned long     regi_size;
   unsigned long     transistors_num;
   unsigned long     gates_num;
   unsigned long     rt_assignement_num;
   unsigned long     rt_condition_num;
   unsigned long     memory_num;
   unsigned long     locvar_num;
   unsigned long     proc_num;
   unsigned long     unused;
   unsigned long     gates_array_num;

   /* delays */

   unsigned long     delays_num;
   unsigned long     delay_unit;
   unsigned long     min_delay;
   unsigned long     max_delay;

   /* initialization table */

   unsigned long     init_num;

   /* C-modules and functions */

   unsigned long     c_modules_num;
   unsigned long     c_functions_num;

   /* maximum parameters */

   unsigned long     max_strength;
   unsigned long     max_size;
   unsigned long     max_fanout;
   unsigned long     max_atpg_level;

   /* link date and time */

   unsigned long     link_date;
   unsigned long     link_time;
   unsigned long     c_link_date;

   /* MOZART internal information */

   unsigned long     vdd_offset;
   unsigned long     gnd_offset;
   unsigned long     static_size;

   /* flags */

   unsigned long     atpg_flag;
   unsigned long     sim_flag;
   unsigned long     gate_array_flag;
   unsigned long     case_sensitive_flag;

   /* topological data structure sizes */

   unsigned long     tot_elements_num;
   unsigned long     tot_fanin_num;
   unsigned long     tot_fanout_num;
   unsigned long     tot_bidir_num;

   /* averages values */

   unsigned long     nodes_fo;
   unsigned long     nodes_fi;
   unsigned long     nodes_conn;
   unsigned long     wires_fo;
   unsigned long     wires_fi;
   unsigned long     regi_fo;
   unsigned long     regi_fi;
   unsigned long     gates_fo;
   unsigned long     gates_fi;
   unsigned long     rt_a_fo;
   unsigned long     rt_a_fi;
   unsigned long     rt_c_fo;
   unsigned long     rt_c_fi;



   /* beginning of executable part */


   /* read the header information from the network file */

   fscanf (net_file, "%s", linker_version);

   /* check the linker version */

   if (strcmp (linker_version, LINKER_VERSION))
   {
      printf ("The linker version is not compatible with me.\n");
      good_bye (TRUE, TRUE);
   }


   fscanf (net_file, "%s", circuit_name);

   fscanf (net_file, "%ld", &sym_tab_offset);
   fscanf (net_file, "%ld", &mod_tab_offset);

   fscanf (net_file, "%ld", &warnings);

   fscanf (net_file, "%ld", &p_i_num);
   fscanf (net_file, "%ld", &p_i_size);
   fscanf (net_file, "%ld", &p_o_num);
   fscanf (net_file, "%ld", &p_o_size);
   fscanf (net_file, "%ld", &p_b_num);
   fscanf (net_file, "%ld", &p_b_size);
   fscanf (net_file, "%ld", &constants_num);
   fscanf (net_file, "%ld", &nodes_num);
   fscanf (net_file, "%ld", &wires_num);
   fscanf (net_file, "%ld", &wires_size);
   fscanf (net_file, "%ld", &regi_num);
   fscanf (net_file, "%ld", &regi_size);
   fscanf (net_file, "%ld", &transistors_num);
   fscanf (net_file, "%ld", &gates_num);
   fscanf (net_file, "%ld", &rt_assignement_num);
   fscanf (net_file, "%ld", &rt_condition_num);
   fscanf (net_file, "%ld", &memory_num);
   fscanf (net_file, "%ld", &locvar_num);
   fscanf (net_file, "%ld", &proc_num);
   fscanf (net_file, "%ld", &unused);
   fscanf (net_file, "%ld", &gates_array_num);

   fscanf (net_file, "%ld", &delays_num);
   fscanf (net_file, "%ld", &delay_unit);
   fscanf (net_file, "%ld", &min_delay);
   fscanf (net_file, "%ld", &max_delay);

   fscanf (net_file, "%ld", &init_num);

   fscanf (net_file, "%ld", &unused);
   fscanf (net_file, "%ld", &unused);

   fscanf (net_file, "%ld", &c_modules_num);
   fscanf (net_file, "%ld", &c_functions_num);

   fscanf (net_file, "%ld", &max_strength);
   fscanf (net_file, "%ld", &max_size);
   fscanf (net_file, "%ld", &max_fanout);
   fscanf (net_file, "%ld", &max_atpg_level);

   fscanf (net_file, "%ld", &vdd_offset);
   fscanf (net_file, "%ld", &gnd_offset);

   fscanf (net_file, "%ld", &link_date);
   fscanf (net_file, "%ld", &link_time);
   fscanf (net_file, "%ld", &c_link_date);

   fscanf (net_file, "%ld", &static_size);

   fscanf (net_file, "%ld", &atpg_flag);
   fscanf (net_file, "%ld", &sim_flag);
   fscanf (net_file, "%ld", &gate_array_flag);
   fscanf (net_file, "%ld", &case_sensitive_flag);

   fscanf (net_file, "%ld", &tot_elements_num);
   fscanf (net_file, "%ld", &tot_fanin_num);
   fscanf (net_file, "%ld", &tot_fanout_num);
   fscanf (net_file, "%ld", &tot_bidir_num);

   fscanf (net_file, "%ld", &nodes_fo);
   fscanf (net_file, "%ld", &nodes_fi);
   fscanf (net_file, "%ld", &nodes_conn);
   fscanf (net_file, "%ld", &wires_fo);
   fscanf (net_file, "%ld", &wires_fi);
   fscanf (net_file, "%ld", &regi_fo);
   fscanf (net_file, "%ld", &regi_fi);
   fscanf (net_file, "%ld", &gates_fo);
   fscanf (net_file, "%ld", &gates_fi);
   fscanf (net_file, "%ld", &rt_a_fo);
   fscanf (net_file, "%ld", &rt_a_fi);
   fscanf (net_file, "%ld", &rt_c_fo);
   fscanf (net_file, "%ld", &rt_c_fi);



   /* allocate the delay table and read the delay values */

   delay_table = (DELAY_TABLE_ELEMENT *)
                      get_memory ((unsigned long) delays_num,
                                  (unsigned long) sizeof (DELAY_TABLE_ELEMENT));

   {

      /* local variable */

      unsigned long  delay_counter;
      unsigned long  fall_tmp;
      unsigned long  rise_tmp;


      for (delay_counter = 0; delay_counter < delays_num; delay_counter++)
      {
         fscanf (net_file, "%ld %ld", &rise_tmp, &fall_tmp);
         delay_table[delay_counter].rise = rise_tmp << 1;
         delay_table[delay_counter].fall = fall_tmp << 1;
      }
   }



   /* the C-module section, C-module table and C-function table are
      not yet implemeted */

   if (c_modules_num)
   {
      printf ("C modules handling not allowed.\n");
      good_bye (TRUE, TRUE);
   }

   if (c_functions_num)
   {
      printf ("C functions handling.\n");
      good_bye (TRUE, TRUE);
   }



   /* store the number of network elements in the global variable */

   num_of_elements = tot_elements_num;



   /* allocate the memory for topological data structures */

   topology_vector = (TOPOLOGICAL_ELEMENT *) get_memory
                             ((unsigned long) tot_elements_num,
                              (unsigned long) sizeof (TOPOLOGICAL_ELEMENT));

   if (tot_fanin_num > 0)
      fanin_vector = (FANIN_ELEMENT *) get_memory
                             ((unsigned long) tot_fanin_num,
                              (unsigned long) sizeof (FANIN_ELEMENT));

   if (tot_fanout_num > 0)
      fanout_vector = (FANOUT_ELEMENT *) get_memory
                             ((unsigned long) tot_fanout_num,
                              (unsigned long) sizeof (FANOUT_ELEMENT));


   /* load the topological elements */

   {

      /* local variables */

      unsigned long   te_counter;
      unsigned long   te_attribute;
      unsigned long   te_code;
      unsigned long   te_sim_level;
      unsigned long   te_atpg_level;
      unsigned long   te_dim;
      unsigned long   te_size;
      unsigned long   te_type;
      unsigned long   te_strength;
      unsigned long   te_0_strength;
      unsigned long   te_1_strength;
      unsigned long   te_delay_index;
      unsigned long   te_fo_num;
      unsigned long   te_fi_num;
      unsigned long   te_conn_num;
      unsigned long   fo_offset;
      unsigned long   fi_offset;
      unsigned long   conn_offset;
      unsigned long   link_counter;
      unsigned long   link_type;
      unsigned long   link_index;
      unsigned long   link_bit;
      unsigned long   link_b1c;
      unsigned long   link_b2c;
      unsigned long   link_b1d;
      unsigned long   link_b2d;
      unsigned long   link_item;
      unsigned long   link_pin;
      unsigned long   const_type;
      unsigned char   const_string [MAX_DATA_SIZE];


      /* beginning of executable part */

      fo_offset = fi_offset = conn_offset = 0;


      /* initialize the maximum level number */

      max_level_number = 0;


      /* initialize the maximum link number */

      max_eval_link_number = 0;


      for (te_counter = 0; te_counter < tot_elements_num; te_counter++)
      {
         topology_vector[te_counter].list_head = (MACHINE *) NULL;

         /* read the first record of the topological descriptor and load some
            fields */

         fscanf (net_file, "%ld %ld %ld %ld", &te_attribute, &te_code,
                                              &te_sim_level, &te_atpg_level);

         switch ( (unsigned char) te_attribute)
         {
            case MZ_INTERNAL      : topology_vector[te_counter].position =
                                                                     INTERNAL;
                                    topology_vector[te_counter].detection_point
                                                                      = FALSE;
                                    break;

            case MZ_INPUT         : topology_vector[te_counter].position =
                                                                        INPUT;
                                    topology_vector[te_counter].detection_point
                                                                      = FALSE;
                                    break;

            case MZ_OUTPUT        : topology_vector[te_counter].position =
                                                                       OUTPUT;
                                    topology_vector[te_counter].detection_point
                                                                      = TRUE;
                                    break;

            case MZ_BIDIRECTIONAL : topology_vector[te_counter].position =
                                                                BIDIRECTIONAL;
                                    topology_vector[te_counter].detection_point
                                                                      = FALSE;
                                    break;

            case MZ_PHANTOM       : topology_vector[te_counter].position =
                                                                      PHANTOM;
                                    topology_vector[te_counter].detection_point
                                                                      = FALSE;
                                    break;
         }

         if (max_level_number < te_sim_level)
         {
            max_level_number = te_sim_level;
         }

         topology_vector[te_counter].level = te_sim_level;
         topology_vector[te_counter].fanin_position = 
                                     fanin_vector + fi_offset;
         topology_vector[te_counter].fanout_position = 
                                     fanout_vector + fo_offset;
         topology_vector[te_counter].last_update_time = MAX_UNSIGNED;
         topology_vector[te_counter].stack_link =
                                        (TOPOLOGICAL_ELEMENT *) MAX_UNSIGNED;
         topology_vector[te_counter].appendix = (APPENDIX *) NULL;

#ifdef RESTORE_ENABLED
         topology_vector[te_counter].saved_list_head = (MACHINE *) NULL;
         topology_vector[te_counter].save_time = MAX_UNSIGNED;
#endif

         switch ( (unsigned char) te_code)
         {
            case MZ_NODE:
                     topology_vector[te_counter].type = NODE;

                     /* read the second record of the topological descriptor */

                     fscanf (net_file, "%ld %ld %ld %ld %ld %ld",
                                       &te_dim, &te_size, &te_delay_index,
                                       &te_fo_num, &te_fi_num, &te_conn_num);

                     if (te_dim != 1)
                     {
                        printf ("Vectorial nodes not allowed.\n");
                        good_bye (TRUE, TRUE);
                     }

                     topology_vector[te_counter].data_size = te_dim;
                     topology_vector[te_counter].data_type = SWITCH_MODEL;
                     topology_vector[te_counter].fanout_number = te_fo_num;
                     topology_vector[te_counter].fanin_number = te_fi_num;
                     topology_vector[te_size].driving_capabilities = te_size;

                     break;

            case MZ_WIRE:

                     /* read the second record of the topological descriptor */

                     fscanf (net_file, "%ld %ld %ld %ld %ld %ld",
                                       &te_dim, &te_size, &te_delay_index,
                                       &te_fo_num, &te_fi_num, &te_conn_num);

                     if (te_dim > MAX_DATA_SIZE)
                     {
                        printf ("Unlimited data bit-length not allowed.\n");
                        good_bye (TRUE, TRUE);
                     }

                     topology_vector[te_counter].data_size = te_dim;
                     topology_vector[te_counter].data_type = RT_MODEL;
                     topology_vector[te_counter].fanout_number = te_fo_num;
                     topology_vector[te_counter].fanin_number = te_fi_num;

                     /* is the size really meaningful? */
                     if (te_size == 0)
                     {
                        topology_vector[te_counter].type = WIRE;
                     }
                     else
                     {
                        topology_vector[te_counter].type = MEM_WIRE;
                     }

                     /* link the element to the second creation pass list 
                        to store the strength of its signal read from its 
                        fanin elements */
                     topology_vector[te_counter].stack_link = second_pass;
                     second_pass = &(topology_vector[te_counter]);

                     break;

            case MZ_CONST:
                     topology_vector[te_counter].type = CONSTANT;

                     /* read the second record of the topological descriptor */

                     fscanf (net_file, "%s %ld %ld", const_type, 
                                                     &te_fo_num, &te_conn_num,
                                                     const_string);
                     te_fi_num = 0;
                     te_dim = strlen (const_string);

                     if (te_dim > MAX_DATA_SIZE)
                     {
                        printf ("Unlimited data bit-length not allowed.\n");
                        good_bye (TRUE, TRUE);
                     }

                     topology_vector[te_counter].data_size = te_dim;
                     topology_vector[te_counter].data_type = RT_MODEL;
                     topology_vector[te_counter].fanout_number = te_fo_num;
                     topology_vector[te_counter].fanin_number = te_fi_num;
                     SET_ST_STRENGTH (topology_vector[te_counter]
                                                        .driving_capabilities,
                                      MAX_STRENGTH + 1, MAX_STRENGTH + 1);
                     break;

            case MZ_TRANSISTOR:
                     /* read the second record of the topological descriptor */

                     fscanf (net_file, "%ld %ld", &te_type, &te_strength);
                     te_fo_num = te_fi_num = te_conn_num = 0;

                     topology_vector[te_counter].data_size = 1;
                     topology_vector[te_counter].data_type = RT_MODEL;
                     topology_vector[te_counter].fanout_number = te_fo_num;
                     topology_vector[te_counter].fanin_number = te_fi_num;
                     SET_ST_STRENGTH (topology_vector[te_counter]
                                                        .driving_capabilities,
                                      te_strength, te_strength);

                     switch ( (unsigned char) te_type)
                     {
                        case MZ_DEPLETION:
                               topology_vector[te_counter].type = DEPL_TRANS;
                               break;
                        case MZ_N_CHANNEL:
                               topology_vector[te_counter].type = N_CH_TRANS;
                               break;
                        case MZ_P_CHANNEL:
                               topology_vector[te_counter].type = P_CH_TRANS;
                               break;
                        default: printf ("unknown transistor type");
                                 good_bye (TRUE, TRUE);
                     }

                     break;

            case MZ_GATE:
                     /* read the second record of the topological descriptor */

                     fscanf (net_file, "%ld %ld %ld %ld %ld %ld",
                                      &te_type, &te_delay_index, &te_0_strength,
                                      &te_1_strength, &te_fo_num, &te_fi_num);
                     te_conn_num = 0;
                     te_dim = 1;

                     topology_vector[te_counter].data_size = te_dim;
                     topology_vector[te_counter].data_type = RT_MODEL;
                     topology_vector[te_counter].fanout_number = te_fo_num;
                     topology_vector[te_counter].fanin_number = te_fi_num;
                     SET_ST_STRENGTH (topology_vector[te_counter]
                                                        .driving_capabilities,
                                      te_0_strength, te_1_strength);

                     switch ( (unsigned char) te_type)
                     {
                        case MZ_AND       :
                                 topology_vector[te_counter].type = AND;
                                 break;
                        case MZ_NAND      :
                                 topology_vector[te_counter].type = NAND;
                                 break;
                        case MZ_OR        :
                                 topology_vector[te_counter].type = OR;
                                 break;
                        case MZ_NOR       :
                                 topology_vector[te_counter].type = NOR;
                                 break;
                        case MZ_BUF       :
                                 topology_vector[te_counter].type = BUF;
                                 break;
                        case MZ_INV       :
                                 topology_vector[te_counter].type = INV;
                                 break;
                        case MZ_XOR       :
                                 topology_vector[te_counter].type = XOR;
                                 break;
                        case MZ_XNOR      :
                                 topology_vector[te_counter].type = XNOR;
                                 break;
                        case MZ_TGP       :
                                 topology_vector[te_counter].type = TGP      ;
                                 break;
                        case MZ_TGN       :
                                 topology_vector[te_counter].type = TGN      ;
                                 break;
                        case MZ_MTGP      :
                                 topology_vector[te_counter].type = MTGP     ;
                                 break;
                        case MZ_MTGN      :
                                 topology_vector[te_counter].type = MTGN     ;
                                 break;
                        case MZ_BUS_TRI   :
                                 topology_vector[te_counter].type = BUS_TRI  ;
                                 break;
                        case MZ_BUS_MOS   :
                                 topology_vector[te_counter].type = BUS_MOS  ;
                                 break;
                        case MZ_BUS_UP    :
                                 topology_vector[te_counter].type = BUS_UP   ;
                                 break;
                        case MZ_BUS_DOWN  :
                                 topology_vector[te_counter].type = BUS_DOWN ;
                                 break;
                        case MZ_AND_TRI   :
                                 topology_vector[te_counter].type = AND_TRI  ;
                                 break;
                        case MZ_NAND_TRI  :
                                 topology_vector[te_counter].type = NAND_TRI ;
                                 break;
                        case MZ_OR_TRI    :
                                 topology_vector[te_counter].type = OR_TRI   ;
                                 break;
                        case MZ_NOR_TRI   :
                                 topology_vector[te_counter].type = NOR_TRI  ;
                                 break;
                        case MZ_BUF_TRI   :
                                 topology_vector[te_counter].type = BUF_TRI  ;
                                 break;
                        case MZ_INV_TRI   :
                                 topology_vector[te_counter].type = INV_TRI  ;
                                 break;
                        case MZ_AND_TRIN  :
                                 topology_vector[te_counter].type = AND_TRIN ;
                                 break;
                        case MZ_NAND_TRIN :
                                 topology_vector[te_counter].type = NAND_TRIN;
                                 break;
                        case MZ_OR_TRIN   :
                                 topology_vector[te_counter].type = OR_TRIN  ;
                                 break;
                        case MZ_NOR_TRIN  :
                                 topology_vector[te_counter].type = NOR_TRIN ;
                                 break;
                        case MZ_BUF_TRIN  :
                                 topology_vector[te_counter].type = BUF_TRIN ;
                                 break;
                        case MZ_INV_TRIN  :
                                 topology_vector[te_counter].type = INV_TRIN ;
                                 break;
                        case MZ_AND_OC    :
                                 topology_vector[te_counter].type = AND_OC   ;
                                 break;
                        case MZ_NAND_OC   :
                                 topology_vector[te_counter].type = NAND_OC  ;
                                 break;
                        case MZ_OR_OC     :
                                 topology_vector[te_counter].type = OR_OC    ;
                                 break;
                        case MZ_NOR_OC    :
                                 topology_vector[te_counter].type = NOR_OC   ;
                                 break;
                        case MZ_BUF_OC    :
                                 topology_vector[te_counter].type = BUF_OC   ;
                                 break;
                        case MZ_INV_OC    :
                                 topology_vector[te_counter].type = INV_OC   ;
                                 break;
                        case MZ_AND_OE    :
                                 topology_vector[te_counter].type = AND_OE   ;
                                 break;
                        case MZ_NAND_OE   :
                                 topology_vector[te_counter].type = NAND_OE  ;
                                 break;
                        case MZ_OR_OE     :
                                 topology_vector[te_counter].type = OR_OE    ;
                                 break;
                        case MZ_NOR_OE    :
                                 topology_vector[te_counter].type = NOR_OE   ;
                                 break;
                        case MZ_BUF_OE    :
                                 topology_vector[te_counter].type = BUF_OE   ;
                                 break;
                        case MZ_INV_OE    :
                                 topology_vector[te_counter].type = INV_OE   ;
                                 break;
                        case MZ_SRNOR     :
                                 topology_vector[te_counter].type = SRNOR    ;
                                 break;
                        case MZ_SRNAND    :
                                 topology_vector[te_counter].type = SRNAND   ;
                                 break;
                        case MZ_MUX       :
                                 topology_vector[te_counter].type = MUX      ;
                                 break;


                        /* flip-flops */

                        case MZ_DFF_PET   :
                                 topology_vector[te_counter].type = DFF_PET;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_PET   :
                                 topology_vector[te_counter].type = TFF_PET;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_DFF_PET_C :
                                 topology_vector[te_counter].type = DFF_PET_C;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_PET_C :
                                 topology_vector[te_counter].type = TFF_PET_C;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_DFF_PET_S :
                                 topology_vector[te_counter].type = DFF_PET_S;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_PET_S :
                                 topology_vector[te_counter].type = TFF_PET_S;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_DFF_PET_SC:
                                 topology_vector[te_counter].type = DFF_PET_SC;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_PET_SC :
                                 topology_vector[te_counter].type = TFF_PET_SC;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_DFF_NET   :
                                 topology_vector[te_counter].type = DFF_NET;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_NET   :
                                 topology_vector[te_counter].type = TFF_NET;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_DFF_NET_C :
                                 topology_vector[te_counter].type = DFF_NET_C;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_NET_C :
                                 topology_vector[te_counter].type = TFF_NET_C;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_DFF_NET_S :
                                 topology_vector[te_counter].type = DFF_NET_S;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_NET_S :
                                 topology_vector[te_counter].type = TFF_NET_S;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_DFF_NET_SC:
                                 topology_vector[te_counter].type = DFF_NET_SC;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_TFF_NET_SC :
                                 topology_vector[te_counter].type = TFF_NET_SC;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_JKFF_PET    :
                                 topology_vector[te_counter].type = JKFF_PET;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_JKFF_PET_C  :
                                 topology_vector[te_counter].type = JKFF_PET_C;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_JKFF_PET_S  :
                                 topology_vector[te_counter].type = JKFF_PET_S;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 2].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_JKFF_PET_SC  :
                                 topology_vector[te_counter].type = JKFF_PET_SC;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset + 2].attribute = 
                                 fanin_vector[fi_offset + 3].attribute = 
                                                                 NON_TRIGGERING;
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                 fanin_vector[fi_offset + 4].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_LATCH_HIGH  :
                                 topology_vector[te_counter].type = LATCH_HIGH;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 TRIGGERING;
                                 break;

                        case MZ_LATCH_LOW  :
                                 topology_vector[te_counter].type = LATCH_LOW;

                                 /* set the topological link attributes */
                                 fanin_vector[fi_offset].attribute = 
                                 fanin_vector[fi_offset + 1].attribute = 
                                                                 TRIGGERING;
                                 break;

                        default           :
                           printf ("Unknown gate or register type.\n");
                           good_bye (TRUE, TRUE);
                     }
                     break;

            default: printf ("Unknown elements in the network.\n");
                     good_bye (TRUE, TRUE);

         }  /* end of switch (te_code) */


#ifdef VECTORIAL
         /* for debug purpose, the data size of each element may be forced 
            to MAX_DATA_SIZE */
         topology_vector[te_counter].data_size = MAX_DATA_SIZE;
#endif


         /* read and store the links */

         if (max_eval_link_number < (te_fi_num + te_conn_num))
         {
            max_eval_link_number = te_fi_num + te_conn_num;
         }


         /* fanout list */

         for (link_counter = 0; link_counter < te_fo_num; link_counter++)
         {
            fscanf (net_file, "%ld %ld", &link_type, &link_index);

            /* NOTE: some fanout links, e.g., the non triggering ones, 
                     should be hidden, to avoid the propagation of the 
                     activity on them. However, not enough information are 
                     generated by the linker to do this. */

            fanout_vector[fo_offset].link = topology_vector + link_index;
            fanout_vector[fo_offset].attribute = VOID;
            fo_offset++;

            if (te_code == MZ_RT_ASSIGNEMENT)
            {
               /* for now, ignore the information */

               fscanf (net_file, "%ld %ld %ld %ld", &link_b1c, &link_b2c,
                                                    &link_b1d, &link_b2d);
            }
            else
            {
               /* for now, ignore the information */

               fscanf (net_file, "%ld %ld", &link_pin, &link_item);
            }
         }


         /* fanin list */

         for (link_counter = 0; link_counter < te_fi_num; link_counter++)
         {
            fscanf (net_file, "%ld %ld %ld %ld", &link_type, &link_index,
                                                 &link_bit, &link_item);

            if (link_bit != 0)
            {
               printf ("Bit extraction not allowed.\n");
               good_bye (TRUE, TRUE);
            }

            fanin_vector[fi_offset].link = topology_vector + link_index;
            fi_offset++;
         }


         /* the computational appendix for now cannot be loaded */

         if ((te_code == MZ_RT_ASSIGNEMENT) || (te_code == MZ_RT_CONDITION))
         {
            printf ("RT assignements and conditions not allowed.\n");
            good_bye (TRUE, TRUE);
         }

      }  /* end of one element loading */

   }  /* end of all the elements loading */


   /* start the second pass to store the strength in the wire descriptors */

   while (second_pass != NULL)
   {
      unsigned long        link_cnt;
      FANIN_ELEMENT        *fi_link;
      unsigned char        driving_capabilities;
      TOPOLOGICAL_ELEMENT  *next_second_pass;

      /* beginning of executable part */

      /* read the strength of all the elements writing on the wire */

      if (second_pass->fanin_number == 0)
      {
         driving_capabilities = 0;
      }
      else
      {
         fi_link = second_pass->fanin_position;
         driving_capabilities = fi_link->link->driving_capabilities;
      }

      for (link_cnt = 1; link_cnt < second_pass->fanin_number; link_cnt++)
      {
         if (driving_capabilities != fi_link->link->driving_capabilities)
         {
            printf ("Different strength on a single wire. Max is forced.\n");
            SET_ST_STRENGTH (driving_capabilities, 
                             MAX_STRENGTH + 1, MAX_STRENGTH + 1);
            break;
         }
         fi_link++;

      }  /* end of loop on fanin */

      second_pass->driving_capabilities = driving_capabilities;

      next_second_pass = second_pass->stack_link;
      second_pass->stack_link = (TOPOLOGICAL_ELEMENT *) MAX_UNSIGNED;
      second_pass = next_second_pass;

   }



   /* initialize VDD and GND if present */

   {

      /* beginning of executable part */

      if (vdd_offset != MAX_UNSIGNED)
      {
         /* modify the element position code */
         topology_vector[vdd_offset].position = VDD;
      }

      if (gnd_offset != MAX_UNSIGNED)
      {
         /* modify the element position code */
         topology_vector[gnd_offset].position = GND;
      }
   }



   /* load the initialization section.
      Note that since the initialization events are inserted in the
      time queue, it must be initialized before this routine is run. */

   {

      /* local variables */

      unsigned long  init_values_counter;
      unsigned long  init_element_counter;
      unsigned char  init_string [MAX_DATA_SIZE];
      unsigned long  init_internal_value;
      unsigned long  init_index;


      /* beginning of executable part */

      for (init_values_counter = 0; init_values_counter < init_num;
                                                       init_values_counter++)
      {
         /* read the initialization value and the number of elements
            to be initialized */

         fscanf (net_file, "%s %ld", init_string, &init_element_counter);

         for ( ; init_element_counter > 0; init_element_counter--)
         {
            /* read the index of the element to be initialized. */

            fscanf (net_file, "%ld", &init_index);

            /* ignore the information */

         }
      }
   }



   return;

}  /* end of load_net */
